/*
 * © 2011 by YOUSURE Tarifvergleich GmbH
 * Author: Philipp Dunkel <p.dunkel@durchblicker.at>
 */

var levels = [ "ERROR", "WARN", "INFO", "DEBUG" ];

var baselevel = (parseInt(process.env.NODE_DEBUG, 16) & 0x4) ? 3 : 0;

var argsarray = function(args) {
	var res = [];
	for (var idx=0; idx < args.length; idx++) {
		res.push(args[idx]);
	}
	return res;
};

var makeLines = function(args) {
	var result = [];
	for (var idx=0; idx<args.length; idx++) {
		switch(typeof args[idx]) {
			case "string": result.push(args[idx]); break;
			case "object": {
				if (args[idx] === null) {
					// Nothing
				} else if (args[idx].constructor === Array) {
					result = result.concat(makeLines(args[idx]));
				} else if (args[idx].constructor === Date) {
					result.push("[object Date] "+JSON.stringify(args[idx]));
				} else if (Buffer.isBuffer(args[idx])) {
					result.push("[object Buffer] ("+args[idx].length+") "+args[idx].toString("utf-8").substring(0,30));
				} else {
					result.push("[object Object] "+JSON.stringify(args[idx]));
				}
				break;
			}
			case "undefined": break;
			default : result.push("["+(typeof args[idx])+"] "+String(args[idx])); break;
		}
	}
	return result;
};
var debug = function() {
	var args = argsarray(arguments);
	var msglevel = args.shift();
	if (msglevel > this.level) {
		/*
		if (!msglevel && ("function" === typeof global.log)) {
			args.unshift("ERROR");
			global.log.apply(global, args);
		}
		*/
		return;
	};
	var tsl = '['+JSON.stringify(new Date()).replace(/^"|Z"$/g,'').replace(/T/g,' ')+'] ';
	args = makeLines(args);
	for (var idx=0; idx<args.length; idx++) {
		console.log(tsl+levels[msglevel]+'('+this.source+')'+((args.length > 1) ? (' '+(idx+1)+' of '+args.length) : '')+': '+(idx ? "\t" : "")+args[idx]);
	}
};

var setup = function(src, lvl) {
	lvl = lvl ? levels.indexOf(String(lvl).toUpperCase()) : baselevel;
	var source = String(src) || "";
	var level = lvl;
	var dbgfunc = function() {
		var args = argsarray(arguments);
		var msglevel = levels.indexOf(String(args[0]).toUpperCase());
		if (msglevel < 0) {
			args.unshift(levels.length - 1);
		} else {
			args[0] = msglevel;
		}
		debug.apply({ level:level, source:source }, args);
	};
	for (var idx=0; idx<levels.length; idx++) {
		dbgfunc[levels[idx]] = levels[idx];
	}
	return dbgfunc;
};

module.exports = setup;
