/*
** Author: Philipp Dunkel <p.dunkel@durchblicker.at>
** Author: Petra Ollram <p.ollram@durchblicker.at>
** © 2011 by YOUSURE Tarifvergleich GmbH
** 
** Permission is hereby granted, free of charge, to any person obtaining a copy
** of this software and associated documentation files (the "Software"), to deal
** in the Software without restriction, including without limitation the rights
** to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
** copies of the Software, and to permit persons to whom the Software is
** furnished to do so, subject to the following conditions:
** 
** The above copyright notice and this permission notice shall be included in
** all copies or substantial portions of the Software.
** 
** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
** IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
** FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
** AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
** LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
** OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
** THE SOFTWARE.
** 
*/

var https = require('https');
var querystring = require("querystring");
var urlparser = require("url");
var fs = require("fs");
var utils = require("util");

var debug = require("./debug")("salesforce.js");

var SFServiceURL = "/services/data/v20.0";

var SF = function (username, password, credential, clientid, clientsecret) {
	this.username = String(username);
	this.password = String(password)+String(credential);
	this.client_id = String(clientid);
	this.client_secret = String(clientsecret);
	this.keepAlive = true;
	this.serviceURL = SFServiceURL;
	this.serviceInfo();
};

SF.serviceInfo = function(opts, cb) {
	if ((!cb) && ("function" === typeof opts)) {
		cb = opts;
		opts = { host:"login.salesforce.com" };
	}
	opts = opts || {};
	var that = this;
	var options = {
	  host: String(opts.host || "login.salesforce.com"),
	  port: 443,
	  path: '/services/data?',
	  method: 'GET'
	};
	if (options.keepAlive) options.headers = { "connection":"keep-alive" };
	debug("SERVICE: "+JSON.stringify(options));
	var request = https.request(options, function(response) {
		var data = new ScaleBuffer();
		response.on("data",function(chunk) { data.write(chunk); });
		response.on("end", function() {
			if (response.statusCode > 299) {
				if ("function" === typeof cb) cb.call(that, response.statusCode, undefined);
				return;
			}
			try {
				data = JSON.parse(data.toString("utf-8"));
				debug("SERVICE(response): "+JSON.stringify(data));
				if ("function" === typeof cb) { cb.call(that, undefined, data); }else debug("No CB 1");
			} catch(ex) {
				if ("function" === typeof cb) { cb.call(that, ex, undefined); }else debug("No CB 2");
				return;
			}
		});
	});
	request.end();
};
SF.prototype.serviceInfo = function(cb) {
	var that = this;
	SF.serviceInfo({ host:this.host, keepAlive:this.keepAlive }, function(err, data) {
		if (!err && data) {
			that.services = data;
			that.serviceURL = that.services[that.services.length-1].url;
		} 
		if (cb && ("function" === typeof cb)) { cb.call(that, err, data); } else debug("No CB 3");
	});
};

SF.prototype.login = function (host, cb) {
	if ((!cb) && ("function" === typeof host)) {
		cb = host;
		host = "login.salesforce.com";
	}
	var that = this;
	var options = {
	  host: String(host),
	  port: 443,
	  path: '/services/oauth2/token?'+querystring.stringify({
			"grant_type":"password",
			"client_id":this.client_id,
			"client_secret":this.client_secret,
			"username":this.username,
			"password":this.password,
			"format":"json"
		}),
	  method: 'POST',
		headers:{
			"Content-Length":0
		}
	};
	debug("LOGIN: "+JSON.stringify(options));
	var request = https.request(options, function(response) {
		if (response.statusCode > 299) {
			debug("Login Error "+response.statusCode);
			if ("function" === typeof cb) cb(response.statusCode, undefined);
		}
		var data = new ScaleBuffer();
		response.on("data",function(chunk) { data.write(chunk); });
		response.on("end", function() { 
			if (response.statusCode > 299) {
				if ("function" === typeof cb) cb.call(that, response.statusCode, undefined);
				return;
			}
			try {
				data = JSON.parse(data.data.toString("utf-8"));
			} catch(ex) {
				if ("function" === typeof cb) cb.call(that, ex, undefined);
				return;
			}
			debug("LOGIN(response): "+JSON.stringify(data));
			that.host = urlparser.parse(data.instance_url).hostname;
			that.token = data.access_token;
			if ("function" === typeof cb) cb.call(that, undefined, data);
			return;
		});
	});
	request.end();
};
SF.prototype.makeURL = function(command, params) {
	if (command.constructor !== Array) command=[ command ];
	command.unshift(this.serviceURL);
	var url = command.join("/");
	if (params) {
		url += "?"+querystring.stringify(params);
	}
	return url;
};

SF.prototype.request = function(url, method, data, cb) {
	if (!cb) {
		if ("function" === typeof data) {
			cb = data;
			data = undefined;
		} else if ("function" === typeof method) {
			cb = method;
			data = undefined;
			method = undefiend;
		}
	}
	if (method && !data && ("string" === typeof method) && ([ "GET", "PUT", "POST", "PATCH", "DELETE" ].indexOf(method.toUpperCase()) === -1)) {
		data = method;
		method = undefined;
	}

	var that = this;
	url = String(url);
	method = String(method || "GET").toUpperCase();
	if ((!cb) && ("function" === typeof data)) {
		cb = data;
		data = undefined;
	}
	if (data) {
		if (!Buffer.isBuffer(data)) {
			data = new Buffer(data);
		}
	}
	var exec = function(err, login) {
		if (err) {
			debug("Login-Error: "+err);
			if ("function" === typeof cb) cb.call(that, 403, undefined);
			return;
		}
		var options = {
		  host: that.host,
		  port: 443,
		  path: url,
		  method: method,
			headers:{ "Authorization":"OAuth "+that.token }
		};
		if (that.keepAlive) options.headers["connection"] = "keep-alive";
		options.headers["content-type"] = "application/json";
		if (data) {
			options.headers["content-length"] = data.length;
			options.headers["expect"] = "100-continue";
		}
		debug("REQUEST: "+method+" https://"+(that.host || "login.salesforce.com")+url, JSON.stringify(options.headers), data?data.toString("utf-8"):"");
		
		var request = https.request(options, function(response) {
			debug("Got Response: ("+options.path+")"+response.statusCode);
			data = new ScaleBuffer();
			response.on("data",function(chunk) { data.write(chunk); });
			response.on("end", function() {
				if (response.statusCode === 401) {
					that.token = undefined;
					that.request(url, method, data, cb);
					return;
				}
				if (response.statusCode === 204) {
					if ("function" === typeof cb) cb.call(that, undefined, { errors:[], success:true });
					return;
				}
				if (response.headers['content-type'].match(/^application\/json/)) {
					try {
						data = JSON.parse(data.data.toString("utf-8"));
						if ("function" === typeof cb) cb.call(that, response.statusCode > 299 ? response.statusCode : undefined, data);
						return;
					} catch (ex) {
						if ("function" === typeof cb) cb.call(that, response.statusCode > 299 ? response.statusCode : ex, undefined);
						return;
					}
				} else {
					if ("function" === typeof cb) cb.call(that, response.statusCode > 299 ? response.statusCode : undefined, { 'type':response.headers['content-type'], 'content':data.data });
				}
			});
		});
		request.on("error", function(err) {
			debug(debug.ERROR, err);
			if ("function" === typeof cb) cb.call(that, err, undefined);
		});
		if (data) {
			request.on("continue", function() {
				debug("Continue ("+options.path+")");
				request.end(data.length ? data : undefined);
			});
		} else {
			debug("End ("+options.path+")");
			request.end();
		}
	};
	if (this.token) {
		exec(undefined, {});
	} else {
		this.login(exec);
	}
};

SF.prototype.close = function(cb) {
	var that = this;
	this.keepAlive = false;
	this.serviceInfo(function(err, data) {
		that.serviceURL = SFServiceURL;
		delete that.token;
		delete that.services;
		that.keepAlive = true;
		if (cb && ("function" === typeof cb)) cb.call(that);
	});
};

SF.prototype.describe = function(objectclass, cb) {
	if ((!cb) && ("function" === typeof objectclass)) {
		cb = objectclass;
		objectclass = undefined;
	}
	debug("describe("+(objectclass || '')+")");
	this.request(this.makeURL(objectclass ? [ 'sobjects', objectclass, 'describe' ] : [ 'sobjects' ]), 'GET', cb);
};
SF.prototype.queryObjects = function(query, cb) {
	debug('queryObjects('+query+')');
	this.request(this.makeURL('query',{ q:query }), 'GET', cb);
};
SF.prototype.searchObjects = function(query, cb) {
	debug('searchObjects('+query+')');
	this.request(this.makeURL('search',{ q:query }), 'GET', cb);
};

SF.prototype.createObject = function(objectclass, object, cb) {	
	debug("createObject("+objectclass+")");
	this.request(this.makeURL([ 'sobjects', objectclass ]), "POST", JSON.stringify(object), cb);
};
SF.prototype.fetchBlobField = function(objectclass, id, field, cb) {
	debug("fetchBlobField("+objectclass+":"+id+":"+field+")");
	var that = this;
	this.request(this.makeURL([ 'sobjects', objectclass, id, field ]), 'GET', function(err, data) {
		if (!err) {
			data.id = id;
		}
		if (cb && ("function" === typeof cb)) cb.call(that, err, data);
	});
};

SF.prototype.fetchObject = function(objectclass, id, fields, cb) {
	if ((!cb) && ("function" === typeof fields)) {
		cb = fields;
		fields = undefined;
	}
	if ("object" !== typeof fields) fields=undefined;
	if (fields && (fields.constructor !== Array)) {
		var tmp = [];
		var item;
		for (item in fields) {
			if (fields.hasOwnProperty(item)) tmp.push(item);
		}
		fields = tmp.length ? tmp : undefined;
	}
	debug("fetchObject("+objectclass+":"+id+", "+(fields ? JSON.stringify(fields) : '[]')+")");
	this.request(this.makeURL([ 'sobjects', objectclass, id ], fields ? { fields:fields.join("," ) } : undefined), 'GET', cb);
};
SF.prototype.updateObject = function(objectclass, id, object, cb) {
	debug("updateObject("+objectclass+":"+id+")");
	this.request(this.makeURL([ 'sobjects', objectclass, id ]), "PATCH", JSON.stringify(object), cb);
};
SF.prototype.upsertObject = function(objectclass, object, field, cb) {
	debug("upsertObject("+objectclass+":"+field+"="+object[field]+")");
	this.request(this.makeURL([ 'sobjects', objectclass, field, object[field] ]), "PATCH", JSON.stringify(object), cb);
};
SF.prototype.deleteObject = function(objectclass, id, cb) {
	debug("deleteObject("+objectclass+":"+id+")");
	this.request(this.makeURL([ 'sobjects', objectclass, id ]), "DELETE", cb);
};

SF.prototype.fetchExternalObject = function(objectclass, field, value, cb) {
	debug("fetchExternalObject("+objectclass+":"+field+"="+value+")");
	
	this.request(this.makeURL([ 'sobjects', objectclass, field, value ]), 'GET', cb);
};
SF.prototype.updateExternalObject = function(objectclass, object, field, cb) {
	debug("updateExternalObject("+objectclass+":"+field+"="+object[field]+")");
	this.request(this.makeURL([ 'sobjects', objectclass, field, object[field] ]), "PATCH", JSON.stringify(object), cb);
};
SF.prototype.upsertExternalObject = SF.prototype.upsertObject;
SF.prototype.deleteExternalObject = function(objectclass, field, value, cb) {
	debug("deleteExternalObject("+objectclass+":"+field+"="+object[field]+")");
	this.request(this.makeURL([ 'sobjects', objectclass, field, object[field] ]), "DELETE", cb);
};


/*
Oh I wish it wasn't so, but Node currently cannt http-send a lot of data. If it does the request simply never ends
So I had to do it via curl. (Change this as soon as node is fixed)
*/
var spawn = require("child_process").spawn;
SF.prototype.createAttachment = function(parentid, name, content, contenttype, cb) {
	var that = this;
	var exec = function(err, data) {
		debug("Adding Attachment");
		var child = spawn("curl",[
			"https://"+that.host+that.makeURL([ 'sobjects', "Attachment" ]),
			"-X", "POST", "-s",
			"-H", "Authorization: OAuth "+that.token,
			"-H", "Content-Type: application/json",
			"-d", "@/dev/stdin",
			"-D", "/dev/stderr"
		]);

		if (!child) {
			if (cb && ("function" != typeof cb)) cb.call(that, "Execution Failed", undefined);
			return;
		}

		data = new ScaleBuffer();
		child.stdout.on("data", function(chunk) { data.write(chunk); });
		var hdat = new ScaleBuffer();
		child.stderr.on("data", function(chunk) { hdat.write(chunk); });

		child.on("exit",function(code) {
			if (code) {
				code = { exit:code, error:hdat.data.toString("utf-8") };
				debug("Child-Exit: ", JSON.stringify(code));
				if (cb && ("function" != typeof cb)) cb.call(that, code, undefined);
				return;
			}
			try {
				data = JSON.parse(data.data.toString("utf-8"));
			} catch(ex) {
				data = data.data.toString("utf-8");
			}
			hdat = hdat.data.toString("utf-8");
			hdat = hdat.replace(/^HTTP\/\d+\.\d+\s+100\s+(.|\r|\n)+HTTP\//,"HTTP/").replace(/HTTP\/\d+\.\d+\s+(\d+)\s+.*/,"$1");

			if (isNaN(parseInt(hdat, 10))) {
				debug("Error(1): "+hdat+" ("+cb+")");
				if (cb && ("function" === typeof cb)) cb.call(that, hdat, data);
				return;
			}
			hdat = parseInt(hdat,10);
			if (hdat > 299) {
				debug("Error(2): "+hdat+" ("+cb+")");
				if (cb && ("function" === typeof cb)) cb.call(that, hdat, data);
				return;
			}
			debug("No Error ("+hdat+")");
			if (cb && ("function" === typeof cb)) cb.call(that, undefined, data);
		});

		var obj = JSON.stringify({
			"ParentId":parentid,
			"ContentType":contenttype,
			"Name":name,
			"Body":"<<BODY>>"
		}).split("<<BODY>>",2);
		debug("Hmmm");
		child.stdin.write(obj[0],"utf-8");
		switch (typeof content) {
			case "object" : {
				if (Buffer.isBuffer(content)) {
					child.stdin.write(content.toString("base64"),"utf-8");
					child.stdin.write(obj[1],"utf-8");
					child.stdin.end();
				} else {
					debug("Stream-Content");
					content.on("end", function() {
						debug("Stream-Content Complete");
						child.stdin.write(obj[1],"utf-8");
						child.stdin.end();
					});
					content.pipe(child.stdin, { end: false });
					debug("Stream-Content Sent");
				}
				break;
			}
			default : {
				child.stdin.write((new Buffer(content)).toString("base64"),"utf-8");
				child.stdin.write(obj[1],"utf-8");
				child.stdin.end();
			}
		}
	};
	if (this.token) {
		exec(undefined, {});
	} else {
		this.login(exec);
	}
};
SF.prototype.attachStream = function(parentid, name, stream, contenttype,  cb) {
	if ((!cb) && ("function" === typeof contenttype)) {
		cb=contenttype;
		contenttype = undefined;
	}
	this.createAttachment(parentid, name, stream, contenttype || "application/octet-stream", cb);
};
SF.prototype.attachFile = function(parentid, name, filename, contenttype, cb) {
	if ((!cb) && ("function" === typeof contenttype)) {
		cb=contenttype;
		contenttype = undefined;
	}
	var that = this;
	fs.readFile(filename, function(err, data) {
		if (err) {
			if (cb && ("function" === typeof cb)) cb.call(that, err, undefined);
			return;
		}
		that.attachStream(parentid, name || filename, data, contenttype, cb);
	});
	
};
SF.prototype.attachBuffer = function(parentid, name, buffer, contenttype, cb) {
	if ((!cb) && ("function" === typeof contenttype)) {
		cb=contenttype;
		contenttype = undefined;
	}
	this.createAttachment(parentid, name, buffer, contenttype || "application/octet-stream", cb);
};

module.exports=SF;

var ScaleBuffer = function() {
	this.data=new Buffer(0);
	this.length=0;
};
ScaleBuffer.prototype.write = function(chunk, encode) {
	switch(typeof chunk) {
		case "string":
		case "number":
		case "boolean":
			chunk = new Buffer(chunk+"",encode);
			break;
		case "object":
			if (Array === chunk.constructor) {
				chunk = new Buffer(chunk);
			}
	}
	var buffer = new Buffer(this.data.length + chunk.length);
	this.data.copy(buffer,0,0);
	try {
		chunk.copy(buffer, this.data.length, 0);
	} catch(ex) {
		console.log(ex.message);
		console.log(require("sys").inspect(chunk));
		throw(ex);
	}
	
	this.length = buffer.length;
	return this.data=buffer;
};
ScaleBuffer.prototype.read = function(bytes) {
	bytes = (bytes < this.data.length) ? bytes : this.data.length;
	var res = this.data.slice(0,bytes);
	this.data = this.data.slice(bytes,this.data.length);
	this.length = this.data.length;
	return res;
};

ScaleBuffer.prototype.end = function(chunk, encode) {
	if (chunk) this.write(chunk, encode);
	return this.data;
};
ScaleBuffer.prototype.toString = function(encode) {
	return this.data.toString(encode);
};
ScaleBuffer.prototype.getData = function() {
	return this.data;
};
ScaleBuffer.prototype.toJSON = function() {
	return this.toString("base64");
};
